#include <stdio.h>
#include <stdlib.h>
#include <const.h>
#include <time.h>
#include <errno.h>
#include <modes.h>
#include <types.h>
#include <signal.h>
#include <cglob.h>
#include <INET/socket.h>
#include <INET/in.h>
#include <INET/netdb.h>
#include "os9types.h"

#include "pampad.h"

FILE *log;
int Debug = 0;

void prt_error( char *s) {
    if (Debug) {
	fprintf(log, "%s - Error #%d\n", s, errno);
	fflush(log);
    }	
}

int main(int argc, char **argv) {

path_id fd, path;
char *server[N_SERV] = {"statd&", "confd&", "monitd&", "muond&", "eventd&"};
int sock[N_SERV], newsock;
unsigned short port[N_SERV] = {STATUS_PORT, CONFIG_PORT, MONITOR_PORT, MUON_PORT, EVENT_PORT};
struct sockaddr_in sock_name = {AF_INET};
struct sockaddr_in from;
int namelength = sizeof(struct sockaddr_in);
int i,j;
int res;
fd_set sready;

/* Handle command line options (only -d: enable debug) */
while((argc > 1) && (argv[1][0] == '-')) {
    switch(argv[1][1]) {
	case 'd':     
	Debug = 1;
	break;
	default:
	Debug = 0;
	break;
    }
    argc--;
    argv++;
}

if (Debug) log = fopen("pampad.log","w");

/* REDIRECT STDIN, STDOUT & STDERR TO THE NULL DEVICE */
_os_open("/nil", S_IREAD | S_IWRITE, &fd);
_os_close(0);
_os_dup(fd, &path);
_os_close(1);
_os_dup(fd, &path);
_os_close(2);
_os_dup(fd, &path);

for (i=0; i<N_SERV; i++) {
    
    /* GET A SOCKET in sock[] (gets file descriptors) */
    if ((sock[i] = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
	prt_error ("Socket");
	exit(0);
    }
    
    /* BIND AN ADDRESS TO SOCKET in sock_name, until it works! */
    sock_name.sin_port = htons(port[i]);
    if ( bind( sock[i], (struct sockaddr *)&sock_name, 
	    sizeof (sock_name)) != 0 ) {
	fprintf(log, "Bind Error in port %d\n", port[i]);
	fflush(log);
	while (bind(sock[i], (struct sockaddr *)&sock_name, 
		sizeof (sock_name)) != 0) {
    	    fprintf(log,"Bind error again in port %d\n", port[i] );
    	    fflush(log);
	    sleep(2);
	}
    }
    if(Debug) {
	fprintf(log,"Socket binded OK\n");
	fflush(log);
    }	
    /* LISTEN TO THE SOCKET for incoming connections */
    if ( listen( sock[i], 5 ) < 0 ) {
	prt_error ("Listen");
	exit(-1);
    }
    if(Debug) fprintf(log,"Listen on port %d OK\n", port[i] );
    fflush(log);
} /* end of for */

FD_ZERO(&sready);
for(;;) {
    for (i=0; i<N_SERV ; i++) {
	FD_SET(sock[i],&sready);
    }	
    
    /* SELECT over the sockets, wait forever*/
    if((res=select(sock[N_SERV-1]+1,&sready,(void *)0,(void *)0,(void *)0)) > 0) {

    for (i=0; i<N_SERV ; i++) {
	/* TEST sock[i] */
	if(FD_ISSET(sock[i],&sready) ) {
	    
	    /* ACCEPT CONNECTIONS in sock[i] file descriptor */
    	    if((newsock = accept(sock[i], 
			(struct sockaddr *)&from, &namelength)) == -1) {
		prt_error("Accept");
		exit(-1);
    	    }
	    if(Debug) { 
		fprintf(log,"Connection from host %s, port %d, socket %d\n",
	    	    inet_ntoa(from.sin_addr), ntohs(from.sin_port), newsock);
      		fprintf(log,"Launching %s for client %s\n", server[i], 
	    	    inet_ntoa(from.sin_addr));
		    fflush(log);
	    }      			
    	
	    /* REDIRECT STDIN, STDOUT & STDERR TO THE SOCKET & FORK */
		    _os_close(0);
		    _os_dup(newsock, &path);
		    _os_close(1);
		    _os_dup(newsock, &path);
		    _os_close(2);
		    _os_dup(newsock, &path);
		    _os_close(newsock);
		    
		    system(server[i]);
		    
		    _os_close(0);
		    _os_dup(fd, &path);
		    _os_close(1);
		    _os_dup(fd, &path);
		    _os_close(2);
		    _os_dup(fd, &path);
		    FD_CLR(sock[i],&sready);
    
	} /* end of if */	

    } /* end of for */

} /* end of if */

} /* end of infite for loop */

} /* end of main */

